/*
 * Board initialize code for TANBAC Evaluation board TB0229.
 *
 * (C) Masami Komiya <mkomiya@sonare.it> 2004
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */

#include <common.h>
#include <command.h>
#include <asm/addrspace.h>
#include <asm/inca-ip.h>
#include <pci.h>
#if defined(CFG_JFFS_CUSTOM_PART)
#include <jffs2/jffs2.h>

static struct part_info part[2];
#endif

unsigned long mips_io_port_base = 0;

#if defined(CONFIG_PCI)
static struct pci_controller hose;

extern void init_vr7701_pci (struct pci_controller *hose);
void pci_init_board (void)
{
	init_vr7701_pci(&hose);
}
#endif


extern void eccinit(unsigned int size);

static int tanbac_checkvalid(unsigned int *p)
{
	unsigned int sum;
	int  i, e;

	e = p[1];
	if ( p[0] != 0x54420001 || e > 0x1000 )
		return -1;

	e /= sizeof(*p);

	for ( i = 0, sum = 0xbabababa; i < e; i++ )
	  sum -= p[i];

	return sum;
}

long int initdram(int board_type)
{
	if ( ! tanbac_checkvalid((unsigned int*)0xBFFFF000) ) {
		unsigned int mem = (*((unsigned int*)0xBFFFF150));
		if ( mem > 1024 ) {
			printf("(%d =>)", mem );
			mem = 1024;
		}
		mem <<= 20;
		if (*((unsigned int*)0xBFFFF154) & 0x01 ) {
			printf("ECC ");
			eccinit( mem );
		}
		return mem;
	}
	return get_ram_size (CFG_SDRAM_BASE, 0x10000000);
}


int checkboard (void)
{
	printf("Board: TANBAC TB0290 ");
	printf("(CPU Speed %d MHz)\n",
	       (int)CPU_CLOCK_RATE/1000000);
	return 0;
}


int misc_init_r()
{
	DECLARE_GLOBAL_DATA_PTR;

	int i;
	bd_t *bd = gd->bd;

	if ( tanbac_checkvalid((unsigned int*)0xBFFFF000) ) {
		printf("*** Warning - bad check-sum, using default board config\n" );
	} else {
		for (i = 0; i < 6; ++i) {
			bd->bi_enetaddr[i] = ((unsigned char*)0xBFFFF108)[i];
		}
	}
	return 0;
}

#if defined(CFG_JFFS_CUSTOM_PART)
struct part_info*
jffs2_part_info(int part_num)
{
        if (0 <= part_num && part_num <=1){
                if(part[part_num].usr_priv==(void*)1)
                        return &part[part_num];
                memset(&part[part_num], 0, sizeof(part[0]));
		switch ( part_num ) {
		case 0:
                	part[part_num].offset=(char*)0xBF000000;
                	part[part_num].size=1024*1024*12;
			break;
		case 1:
                	part[part_num].offset=(char*)0xBFC80000;
                	part[part_num].size=0x00340000;
			break;
		}
                /* Mark the struct as ready */
                part[part_num].usr_priv=(void*)1;
                return part;
        }
        return 0;
}
#endif
