/*
 * VR7701 PCI (not PCI-X !)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */

#include <common.h>
#include <pci.h>
#include <asm/addrspace.h>
#include <asm/io.h>

#if defined(CONFIG_PCI)

#define VR7701_PCI_VENDER		KSEG1ADDR(0x18840000)
#define VR7701_PCI_COMMAND		KSEG1ADDR(0x18840004)
#define VR7701_PCI_CALISIZE_LTTIMER	KSEG1ADDR(0x1884000c)

#define VR7701_PCI_PTBADDR0L		KSEG1ADDR(0x18840010)
#define VR7701_PCI_PTBADDR0M		KSEG1ADDR(0x18840014)
#define VR7701_PCI_PTBADDR1L		KSEG1ADDR(0x18840018)
#define VR7701_PCI_PTBADDR1M		KSEG1ADDR(0x1884001c)

/* MAX_LAT, MIN_GNT, INTPIN, INTLINE */
#define VR7701_PCI_MAX_LAT		KSEG1ADDR(0x1884003c)

#define VR7701_PCI_CFGADDR		KSEG1ADDR(0x18840040)
#define VR7701_PCI_CFGDATA		KSEG1ADDR(0x18840044)
#define VR7701_PCI_CFGATTR		KSEG1ADDR(0x18840048)

#define VR7701_PCI_PINTXMASK		KSEG1ADDR(0x18840058)
#define VR7701_PCI_PERRSTAT		KSEG1ADDR(0x18840068)
#define VR7701_PCI_PERRMASK		KSEG1ADDR(0x1884006c)
#define VR7701_PCI_ERRPADRL		KSEG1ADDR(0x18840060)
#define VR7701_PCI_ERRPADRH		KSEG1ADDR(0x18840064)

#define VR7701_PCI_STAT			KSEG1ADDR(0x18840094)
#define VR7701_PCI_SPCMPERRMSG		KSEG1ADDR(0x18840098)

#define VR7701_PCI_PTMEMSEG0		KSEG1ADDR(0x188400a0)
#define VR7701_PCI_PTMEMMSK0		KSEG1ADDR(0x188400a4)
#define VR7701_PCI_PTMEMENSW0		KSEG1ADDR(0x188400a8)

#define VR7701_PCI_PTMEMSEG1		KSEG1ADDR(0x188400b0)
#define VR7701_PCI_PTMEMMSK1		KSEG1ADDR(0x188400b4)
#define VR7701_PCI_PTMEMENSW1		KSEG1ADDR(0x188400b8)

#define VR7701_PCI_PMSEGL		KSEG1ADDR(0x188400c0)
#define VR7701_PCI_PMSEGH		KSEG1ADDR(0x188400c4)
#define VR7701_PCI_PCIXSTAT		KSEG1ADDR(0x18840094)
#define VR7701_PCI_PRST			KSEG1ADDR(0x188400f0)
#define VR7701_PCI_UNITCNT		KSEG1ADDR(0x188400f4)
#define VR7701_PCI_TIMER		KSEG1ADDR(0x188400f8)

#define VR9721_PCI_DRSTCNT		KSEG1ADDR(0x188400ec)

#define PCI_ACCESS_READ  0x06000000
#define PCI_ACCESS_WRITE 0x07000000

/*
 *	Access PCI Configuration Register for VR7701
 */

#define ATTR_READ	0x06000000
#define ATTR_WRITE	0x07000000

static inline void io_sync32( unsigned long io )
{
  // load and store to stack.
	volatile u32 a;
	u32 b = inl(io);
	a = b; // suppress compiler warning
}


static int vr7701_pci_config_access (u32 attr, pci_dev_t dev, int reg,
				     u32 * data)
{
	u32 device;
	u32 idsel = 0;
	u32 err;

	device = PCI_DEV(dev);

	writel(attr|((readl(VR7701_PCI_STAT) & 0x0FF) << 8), VR7701_PCI_CFGATTR);

	writel(0x000307F7, VR7701_PCI_PERRSTAT); //clear error

	if (PCI_BUS(dev) == 0) {
		/* Type 0 Configuration */
		if ( readl(VR7701_PCI_UNITCNT) & 0x1 ) {
			/* PCI mode */
			if (device < 11) {
				return 1;
			}
			idsel = 1UL << device;
			writel( (u32) (idsel | (dev & 0x0700) | (reg & 0xfc)), VR7701_PCI_CFGADDR);
		} else {
			/* PCI-X mode */
			if (device < 16 || device >= 0x1f) {
				return 1;
			}
			idsel = 1UL << device;
			writel( (u32) (idsel | (dev & 0xFF00) | (reg & 0xfc)), VR7701_PCI_CFGADDR);
		}
	} else {
		/* Type 1 Configuration */
		writel( (u32) (dev | (reg & 0xfc) | 1), VR7701_PCI_CFGADDR);
	}

	/* dummy read to wait for all writes are finished */
	io_sync32(VR7701_PCI_VENDER);

	if (attr == PCI_ACCESS_WRITE) {
		writel( *data, VR7701_PCI_CFGDATA );
	} else {
		*data = readl(VR7701_PCI_CFGDATA);
	}

	/* dummy read to wait for all writes are finished */
	io_sync32(VR7701_PCI_VENDER);

	if ((idsel=readl(VR7701_PCI_PERRSTAT)) ) {		/* error! */
		if ( (idsel & ~0x101) || attr == PCI_ACCESS_WRITE ) {
			printf( "PCI:%02x:%02x:%02x[%04x] %s err=%05x ad=%x_%08x split=%08x\n",
				PCI_BUS(dev), PCI_DEV(dev), PCI_FUNC(dev), reg,
				(attr == PCI_ACCESS_WRITE)?"write":"read",
				idsel,
				readl(VR7701_PCI_ERRPADRH), readl(VR7701_PCI_ERRPADRL),
				readl(VR7701_PCI_SPCMPERRMSG)
				);
			printf( "cmd: attr=%08x cfg=%08x\n",
				attr|((readl(VR7701_PCI_STAT) & 0x0FF) << 8),
				(u32) (dev | (reg & 0xfc) | 1)
				);
		}
		writel(idsel, VR7701_PCI_PERRSTAT);
		return 1;
	}

	return (0);
}

static int vr7701_pci_read_config_byte (struct pci_controller* hose,
	pci_dev_t dev, int reg, u8 * val)
{
	u32 data;

	if (vr7701_pci_config_access (PCI_ACCESS_READ, dev, reg, &data)) {
		*val = 0xff;
		return -1;
	}

	*val = (data >> ((reg & 3) << 3)) & 0xff;

	return 0;
}


static int vr7701_pci_read_config_word (struct pci_controller* hose,
	pci_dev_t dev, int reg, u16 * val)
{
	u32 data;

	if (reg & 1) {
		*val = 0xffff;
		return -1;
	}

	if (vr7701_pci_config_access (PCI_ACCESS_READ, dev, reg, &data)) {
		*val = 0xffff;
		return -1;
	}

	*val = (data >> ((reg & 3) << 3)) & 0xffff;

	return 0;
}


static int vr7701_pci_read_config_dword (struct pci_controller* hose,
	pci_dev_t dev, int reg, u32 * val)
{
	u32 data = 0;

	if (reg & 3) {
		*val = 0xffffffff;
		return -1;
	}

	if (vr7701_pci_config_access (PCI_ACCESS_READ, dev, reg, &data)) {
		*val = 0xffffffff;
		return -1;
	}

	*val = data;

	return (0);
}

static int vr7701_pci_write_config_byte (struct pci_controller* hose,
	pci_dev_t dev, int reg, u8 val)
{
	u32 data = 0;

	if (vr7701_pci_config_access (PCI_ACCESS_READ, dev, reg, &data))
		return -1;

	data = (data & ~(0xff << ((reg & 3) << 3))) | (val <<
						       ((reg & 3) << 3));

	if (vr7701_pci_config_access (PCI_ACCESS_WRITE, dev, reg, &data))
		return -1;

	return 0;
}


static int vr7701_pci_write_config_word (struct pci_controller* hose,
	pci_dev_t dev, int reg, u16 val)
{
	u32 data = 0;

	if (reg & 1)
		return -1;

	if (vr7701_pci_config_access (PCI_ACCESS_READ, dev, reg, &data))
		return -1;

	data = (data & ~(0xffff << ((reg & 3) << 3))) | (val <<
							 ((reg & 3) << 3));

	if (vr7701_pci_config_access (PCI_ACCESS_WRITE, dev, reg, &data))
		return -1;

	return 0;
}

static int vr7701_pci_write_config_dword (struct pci_controller* hose,
	pci_dev_t dev, int reg, u32 val)
{
	u32 data;

	if (reg & 3) {
		return -1;
	}

	data = val;

	if (vr7701_pci_config_access (PCI_ACCESS_WRITE, dev, reg, &data))
		return -1;

	return (0);
}


/*
 *	Initialize VR7701 PCIU
 */

static int vr7701_pciu_init (void)
{
  	u32 mode = readl(VR7701_PCI_UNITCNT);
	int speed[4] = { 33, 66, 100, 133 };
	printf( "PCI%s: %dMHz as %s-device\n",
		(mode & 0x1) ? "":"-X",
		speed[(mode>>1)&0x3],
		(mode & 0x10) ? "Host":"Target" );

	if ( readl(VR7701_PCI_VENDER) == 0x01311033 ) { // Vr9721
	  writel(0x00000400, VR9721_PCI_DRSTCNT); // Strength = 4
	}

	if ( mode & 0x10 )
	  writel( 0x00000000, VR7701_PCI_PRST );		/* assert reset# */

	/* STATUS/COMMAND Register */
	writel( 0xFB300006, VR7701_PCI_COMMAND );

	/* LTTIMER Register */
	writel( 0x00004008, VR7701_PCI_CALISIZE_LTTIMER );
		/* MAX LAT Register */
		/* *VR7701_PCI_MAX_LAT = 0xf0f00100; */

	/* PMSEGL Register offset = 0xe0000000, no swap, no prefetchable */ 
	writel( 0xe0000003, VR7701_PCI_PMSEGL );
	/* PMSEGH Register */
	writel( 0xffffffff, VR7701_PCI_PMSEGH );

	/* PTMEMENSW0 Register = no swap */
	writel( 0x00000003, VR7701_PCI_PTMEMENSW0 );
	/* PTMEMENSW1 Register = no swap */
	writel( 0x00000003, VR7701_PCI_PTMEMENSW1 );

	/* PTMEMMSK0 Register, size = 256MB, enable*/
	writel( 0x0fff0001, VR7701_PCI_PTMEMMSK0 );
	/* PTMEMMSK1 Register, disabled */
	writel( 0x00000000, VR7701_PCI_PTMEMMSK1 );

	/* PTBADDR0L Register, window base addr = 0 */
	writel( 0x0000000c, VR7701_PCI_PTBADDR0L );
	/* PTBADDR0M Register */
	writel( 0x00000000, VR7701_PCI_PTBADDR0M );

	/* PTBADDR1L Register */
	writel( 0x0000000c, VR7701_PCI_PTBADDR1L );
	/* PTBADDR1M Register */
	writel( 0x00000001, VR7701_PCI_PTBADDR1M );

	/* PTMEMSEG0 Register, oncip bus base addr = 0  */
	writel( 0x00000000, VR7701_PCI_PTMEMSEG0 );
	/* PTMEMSEG1 Register */
	writel( 0x00000000, VR7701_PCI_PTMEMSEG1 );

	/* PCIXSTAT Register */
	writel(0x00030000, VR7701_PCI_PCIXSTAT); // 133MHz, 64bit dev, dev = 0, func = 0

	writel( 0, VR7701_PCI_PINTXMASK );
	writel( 0x000307f7, VR7701_PCI_PERRMASK); // Enable all pci error
//	writel( 0x00000001, VR7701_PCI_PERRMASK );
	writel( 0x000307F7, VR7701_PCI_PERRSTAT );	/* clear */

	writel( 0x10FFFFFF, VR7701_PCI_TIMER ); // time out

	io_sync32(VR7701_PCI_VENDER);
	if ( mode & 0x10 ) {
		udelay(1500);
		writel( 0x00000001, VR7701_PCI_PRST );		/* deassert reset# */
		writel( mode|0x80000020, VR7701_PCI_UNITCNT );	/* enable arbiter */
		io_sync32(VR7701_PCI_VENDER);
		udelay(1500);
		return 0;
	} else {
		writel( mode|0x80000000, VR7701_PCI_UNITCNT );	/* enable host unit */
		return 1;
	}
}

/*
 *	Initialize Module
 */

void init_vr7701_pci (struct pci_controller *hose)
{
	hose->first_busno = 0;
	hose->last_busno = 0xff;

	if ( vr7701_pciu_init() )	/* Initialize VR7701 PCIU */
		return;			/* pciu is target device */

	/* System memory space mapped over PCI */
	pci_set_region (hose->regions + 0,
			0x00000000,
			0x80000000,
			0x10000000, PCI_REGION_MEM | PCI_REGION_MEMORY);

	/* PCI memory space */
	pci_set_region (hose->regions + 1,
			0xe0000000, 0xe0000000, 0x10000000, PCI_REGION_MEM);

	/* PCI I/O space */
	pci_set_region (hose->regions + 2,
			0x00000000, 0xb8850000, 0x00010000, PCI_REGION_IO);
	hose->region_count = 3;

	hose->read_byte = vr7701_pci_read_config_byte;
	hose->read_word = vr7701_pci_read_config_word;
	hose->read_dword = vr7701_pci_read_config_dword;
	hose->write_byte = vr7701_pci_write_config_byte;
	hose->write_word = vr7701_pci_write_config_word;
	hose->write_dword = vr7701_pci_write_config_dword;

	pci_register_hose (hose);

	hose->last_busno = pci_hose_scan (hose);

	return;
}

#endif
